# -*- coding: utf-8 -*-
"""
Created on Tue Jul 21 10:57:14 2020

@author: rdvan
Linear stability analysis of model 1
"""
from model1_parameters import *
import model1_LSA_functions as mlf
from scipy.optimize import fsolve
import sympy as sym
import numpy as np
from sympy import init_session
init_session() 

# Defining symbolic variables:
X_s, X_i, P, T = sym.symbols('X_s X_i P T')
mu_s, C_s, ep_s, Lambda_s, gamma_s, beta_s, eta_s, alpha_s = sym.symbols('mu C epsilon lambda gamma b eta alpha')
y = sym.symbols('y')
X_t = X_s + X_i

# Define equations
e1 = mu_s*X_s*(1 -X_t/C_s) - ep_s*X_s*P
e2 = ep_s*X_s*P - Lambda_s*X_i
e3 = -ep_s*X_s*P - gamma_s*P + beta_s*Lambda_s*X_i
e4 = -eta_s*T + alpha_s*Lambda_s*X_i

# Solve equations to find equilibrium points   
sol = mlf.solve_and_print([e1, e2, e3, e4], [X_s, X_i, P, T])

# Calculate Jacobian
Jac = mlf.calc_jacobian([e1, e2, e3, e4], [X_s, X_i, P, T])

# Evaluate Jacobian in equilibrium points
# Point 1 (0,0,0,0)
Jac0 = mlf.eval_jacobian(Jac, [X_s, X_i, P], [0, 0, 0])

# Calculate Eigenvalues
evalues0 = Jac0.eigenvals()
print('Eigenvalues of Jacobian evaluated in (0,0,0,0)')
print(evalues0)

# Point 2 (C,0,0,0)
Jac1 = mlf.eval_jacobian(Jac, [X_s, X_i, P], [C_s, 0, 0])


# As intermediate step: calculate determinant to obtain secular equation
# The secular equation sometimes helps in analysing the system manually
det1 = mlf.calc_secular_equation(Jac1)

# Calculate Eigenvalues
print('Eigenvalues of Jacobian evaluated in (C,0,0,0)')
evalues1 = Jac1.eigenvals()

# Print eigenvalues one by one in latex format for report    
mlf.print_eigenval_latex(evalues1)

# Point 3 (..,..,..,..)
Jac2 = mlf.eval_jacobian(Jac, [X_s, X_i, P], [sol[2][0], sol[2][1], sol[2][2]])
Jac2_char = Jac2.charpoly('y')

# Plug in numerical values in Jac2 as eigenvalue calculation is very difficult
Jac2_num = mlf.eval_jacobian(Jac2, [mu_s, C_s, ep_s, Lambda_s, gamma_s, beta_s, eta_s, alpha_s], 
                             [mu_gut, c_gut, ep_gut, lambda_gut, gamma_gut, beta_gut, eta1_gut, alpha_gut])
print(sym.pretty(sym.simplify(Jac2_num)))

# Convert to numpy matrix
Jac2_num = np.array(Jac2_num).astype(np.float64)

# Calculate Eigenvalues
print('Numerical calculation of Eigenvalues of Jacobian evaluated in (..,..,..,..)')
evalues2 = np.linalg.eig(Jac2_num)
print(evalues2[0])


### SYSTEM WITH RESISTANCE
# Defining new symbols
X_r = sym.symbols('X_r')
phi_s = sym.symbols('phi')
X_t = X_s + X_i + X_r

# Define new equations
e1 = mu_s*X_s*(1 -X_t/C_s) - ep_s*X_s*P
e2 = ep_s*X_s*P - Lambda_s*X_i
e3 = -ep_s*X_s*P - gamma_s*P + beta_s*Lambda_s*X_i
e4 = -eta_s*T + alpha_s*Lambda_s*X_i
e5 = mu_s*X_r*(1 -X_t/C_s) + phi_s * X_s

## Symbolic solving takes too long --> numerical solution
# Solve equations to find equilibrium points   
def system_resistance(x, param):
    # Variables
    X_s = x[0]
    X_i = x[1]
    P = x[2]
    T = x[3]
    X_r = x[4]
    
    # parameters
    mu = param[0]                              # bacterial growthrate 1/h
    C = param[1]                                # carrying capacity 1/ml
    ep = param[2]                           # bacterial contact rate constant ml/h
    Lambda = param[3]                      # bacterial lysis death rate 1/h
    gamma = param[4]                         # phage deathrate 1/h
    beta = param[5]                         # virus replication factor
    eta = param[6]                            # toxin degradation rate 1/h
    alpha = param[7]                              # toxin produced per cell - TEMP
    psi = param[8]                   #rate (1/h) of formation of a non suseptible cell - this value is fake
    psi = 1
    
    # equations
    # total bacterial concentration
    X_t = X_s +X_i+ X_r                         
    
    # Define equations
    e1 = mu*X_s*(1 -X_t/C) - ep*X_s*P
    e2 = ep*X_s*P - Lambda*X_i
    e3 = -ep*X_s*P - gamma*P + beta*Lambda*X_i
    e4 = -eta*T + alpha*Lambda*X_i
    e5 = mu*X_r*(1 -X_t/C) + X_s*psi
    
    return [e1, e2, e3, e4, e5]

# Guess for E = (0,0,0,0,0)
x0 = [10, 0.1, 0.1, 0.1, 0.1]
# Guess for E = (0,0,0,0,C)
x0_1 = [100000, 1000, 1000, 1000, c_gut]

#x0_2 = [1000000, 400000, 20000, 10000000, 1000000]

# Solve numerically
eq_point_1 = fsolve(lambda x: system_resistance(x, param_gut), x0)
eq_point_2 = fsolve(lambda x: system_resistance(x, param_gut), x0_1, full_output=1, xtol=1e-15)
#eq_point_3 = fsolve(lambda x: system_resistance(x, param), x0_2)
#print(eq_point_3)

# Calculate Jacobian
Jac_r = mlf.calc_jacobian([e1, e2, e3, e4, e5], [X_s, X_i, P, T, X_r])

# Evaluate Jacobian in equilibrium points
# Point 1 (0,0,0,0,0)
Jac_r0 = mlf.eval_jacobian(Jac_r, [X_s, X_i, P, X_r], [0, 0, 0, 0])
#print(sym.pretty(Jac_r0))

# Calculate Eigenvalues
evalues_r0 = Jac_r0.eigenvals()
print('Eigenvalues of Jacobian evaluated in (0,0,0,0.0)')
print(evalues_r0)

# Evaluate Jacobian in equilibrium points
# Point 1 (0,0,0,0,C)
Jac_r1= mlf.eval_jacobian(Jac_r, [X_s, X_i, P, X_r], [0, 0, 0, C_s])
#print(sym.pretty(Jac_r1))

# Calculate Eigenvalues
evalues_r1 = Jac_r1.eigenvals()
print('Eigenvalues of Jacobian evaluated in (0,0,0,0.0)')
print(evalues_r1)


